@@ -6,14 +6,19 @@ use std::{
     mem::{self, MaybeUninit},
     net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr, SocketAddrV4, SocketAddrV6},
     os::unix::io::AsRawFd,
-    sync::atomic::AtomicUsize,
+    sync::{
+        atomic::{AtomicBool, AtomicUsize},
+        Mutex,
+    },
     time::Instant,
 };
 
-use proto::{EcnCodepoint, Transmit};
 use socket2::SockRef;
 
-use super::{cmsg, log_sendmsg_error, RecvMeta, UdpSockRef, UdpState, IO_ERROR_LOG_INTERVAL};
+use super::{
+    cmsg, log_sendmsg_error, EcnCodepoint, RecvMeta, Transmit, UdpSockRef, UdpState,
+    IO_ERROR_LOG_INTERVAL,
+};
 
 #[cfg(target_os = "freebsd")]
 type IpTosTy = libc::c_uchar;
@@ -26,14 +31,14 @@ type IpTosTy = libc::c_int;
 /// platforms.
 #[derive(Debug)]
 pub struct UdpSocketState {
-    last_send_error: Instant,
+    last_send_error: Mutex<Instant>,
 }
 
 impl UdpSocketState {
     pub fn new() -> Self {
         let now = Instant::now();
         Self {
-            last_send_error: now.checked_sub(2 * IO_ERROR_LOG_INTERVAL).unwrap_or(now),
+            last_send_error: Mutex::new(now.checked_sub(2 * IO_ERROR_LOG_INTERVAL).unwrap_or(now)),
         }
     }
 
@@ -42,12 +47,12 @@ impl UdpSocketState {
     }
 
     pub fn send(
-        &mut self,
+        &self,
         socket: UdpSockRef<'_>,
         state: &UdpState,
         transmits: &[Transmit],
     ) -> Result<usize, io::Error> {
-        send(state, socket.0, &mut self.last_send_error, transmits)
+        send(state, socket.0, &self.last_send_error, transmits)
     }
 
     pub fn recv(
@@ -120,6 +125,12 @@ fn init(io: SockRef<'_>) -> io::Result<()> {
             )?;
         }
     }
+    #[cfg(any(target_os = "freebsd", target_os = "macos", target_os = "ios"))]
+    {
+        if is_ipv4 {
+            set_socket_option(&*io, libc::IPPROTO_IP, libc::IP_DONTFRAG, OPTION_ON)?;
+        }
+    }
     #[cfg(any(target_os = "freebsd", target_os = "macos"))]
     // IP_RECVDSTADDR == IP_SENDSRCADDR on FreeBSD
     // macOS uses only IP_RECVDSTADDR, no IP_SENDSRCADDR on macOS
@@ -130,14 +141,14 @@ fn init(io: SockRef<'_>) -> io::Result<()> {
         }
     }
 
-    // IPV6_RECVPKTINFO is standardized
+    // Options standardized in RFC 3542
     if !is_ipv4 {
         set_socket_option(&*io, libc::IPPROTO_IPV6, libc::IPV6_RECVPKTINFO, OPTION_ON)?;
         set_socket_option(&*io, libc::IPPROTO_IPV6, libc::IPV6_RECVTCLASS, OPTION_ON)?;
-    }
-
-    if !is_ipv4 {
-        set_socket_option(&*io, libc::IPPROTO_IPV6, libc::IPV6_RECVTCLASS, OPTION_ON)?;
+        // Linux's IP_PMTUDISC_PROBE allows us to operate under interface MTU rather than the
+        // kernel's path MTU guess, but actually disabling fragmentation requires this too. See
+        // __ip6_append_data in ip6_output.c.
+        set_socket_option(&*io, libc::IPPROTO_IPV6, libc::IPV6_DONTFRAG, OPTION_ON)?;
     }
 
     Ok(())
@@ -148,10 +159,10 @@ fn send(
     #[allow(unused_variables)] // only used on Linux
     state: &UdpState,
     io: SockRef<'_>,
-    last_send_error: &mut Instant,
+    last_send_error: &Mutex<Instant>,
     transmits: &[Transmit],
 ) -> io::Result<usize> {
-    #[allow(unused_mut)] // only mutable on FeeBSD
+    #[allow(unused_mut)] // only mutable on FreeBSD
     let mut encode_src_ip = true;
     #[cfg(target_os = "freebsd")]
     {
@@ -189,12 +200,15 @@ fn send(
             &mut iovecs[i],
             &mut cmsgs[i],
             encode_src_ip,
+            state.sendmsg_einval(),
         );
     }
     let num_transmits = transmits.len().min(BATCH_SIZE);
 
     loop {
-        let n = unsafe { libc::sendmmsg(io.as_raw_fd(), msgs.as_mut_ptr(), num_transmits as _, 0) };
+        let n = unsafe {
+            sendmmsg_with_fallback(io.as_raw_fd(), msgs.as_mut_ptr(), num_transmits as _)
+        };
         if n == -1 {
             let e = io::Error::last_os_error();
             match e.kind() {
@@ -219,6 +233,12 @@ fn send(
                         }
                     }
 
+                    if e.raw_os_error() == Some(libc::EINVAL) {
+                        // Some arguments to `sendmsg` are not supported.
+                        // Switch to fallback mode.
+                        state.set_sendmsg_einval();
+                    }
+
                     // Other errors are ignored, since they will ususally be handled
                     // by higher level retransmits and timeouts.
                     // - PermissionDenied errors have been observed due to iptable rules.
@@ -241,9 +261,9 @@ fn send(
 
 #[cfg(any(target_os = "macos", target_os = "ios"))]
 fn send(
-    _state: &UdpState,
+    state: &UdpState,
     io: SockRef<'_>,
-    last_send_error: &mut Instant,
+    last_send_error: &Mutex<Instant>,
     transmits: &[Transmit],
 ) -> io::Result<usize> {
     let mut hdr: libc::msghdr = unsafe { mem::zeroed() };
@@ -261,6 +281,7 @@ fn send(
             &mut ctrl,
             // Only tested on macOS
             cfg!(target_os = "macos"),
+            state.sendmsg_einval(),
         );
         let n = unsafe { libc::sendmsg(io.as_raw_fd(), &hdr, 0) };
         if n == -1 {
@@ -289,6 +310,72 @@ fn send(
     Ok(sent)
 }
 
+/// Implementation of `sendmmsg` with a fallback
+/// to `sendmsg` if syscall is not available.
+///
+/// It uses [`libc::syscall`] instead of [`libc::sendmmsg`]
+/// to avoid linking error on systems where libc does not contain `sendmmsg`.
+#[cfg(not(any(target_os = "macos", target_os = "ios")))]
+unsafe fn sendmmsg_with_fallback(
+    sockfd: libc::c_int,
+    msgvec: *mut libc::mmsghdr,
+    vlen: libc::c_uint,
+) -> libc::c_int {
+    let flags = 0;
+
+    #[cfg(not(target_os = "freebsd"))]
+    {
+        let ret = libc::syscall(libc::SYS_sendmmsg, sockfd, msgvec, vlen, flags) as libc::c_int;
+        if ret != -1 {
+            return ret;
+        }
+    }
+
+    // libc on FreeBSD implements `sendmmsg` as a high-level abstraction over `sendmsg`,
+    // thus `SYS_sendmmsg` constant and direct system call do not exist
+    #[cfg(target_os = "freebsd")]
+    {
+        let ret = libc::sendmmsg(sockfd, msgvec, vlen as usize, flags) as libc::c_int;
+        if ret != -1 {
+            return ret;
+        }
+    }
+
+    let e = io::Error::last_os_error();
+    match e.raw_os_error() {
+        Some(libc::ENOSYS) => {
+            // Fallback to `sendmsg`.
+            sendmmsg_fallback(sockfd, msgvec, vlen)
+        }
+        _ => -1,
+    }
+}
+
+/// Fallback implementation of `sendmmsg` using `sendmsg`
+/// for systems which do not support `sendmmsg`
+/// such as Linux <3.0.
+#[cfg(not(any(target_os = "macos", target_os = "ios")))]
+unsafe fn sendmmsg_fallback(
+    sockfd: libc::c_int,
+    msgvec: *mut libc::mmsghdr,
+    vlen: libc::c_uint,
+) -> libc::c_int {
+    let flags = 0;
+    if vlen == 0 {
+        return 0;
+    }
+
+    let n = libc::sendmsg(sockfd, &(*msgvec).msg_hdr, flags);
+    if n == -1 {
+        -1
+    } else {
+        // type of `msg_len` field differs on Linux and FreeBSD,
+        // it is up to the compiler to infer and cast `n` to correct type
+        (*msgvec).msg_len = n as _;
+        1
+    }
+}
+
 #[cfg(not(any(target_os = "macos", target_os = "ios")))]
 fn recv(io: SockRef<'_>, bufs: &mut [IoSliceMut<'_>], meta: &mut [RecvMeta]) -> io::Result<usize> {
     let mut names = [MaybeUninit::<libc::sockaddr_storage>::uninit(); BATCH_SIZE];
@@ -305,12 +392,10 @@ fn recv(io: SockRef<'_>, bufs: &mut [IoSliceMut<'_>], meta: &mut [RecvMeta]) ->
     }
     let msg_count = loop {
         let n = unsafe {
-            libc::recvmmsg(
+            recvmmsg_with_fallback(
                 io.as_raw_fd(),
                 hdrs.as_mut_ptr(),
                 bufs.len().min(BATCH_SIZE) as _,
-                0,
-                ptr::null_mut(),
             )
         };
         if n == -1 {
@@ -352,11 +437,80 @@ fn recv(io: SockRef<'_>, bufs: &mut [IoSliceMut<'_>], meta: &mut [RecvMeta]) ->
     Ok(1)
 }
 
+/// Implementation of `recvmmsg` with a fallback
+/// to `recvmsg` if syscall is not available.
+///
+/// It uses [`libc::syscall`] instead of [`libc::recvmmsg`]
+/// to avoid linking error on systems where libc does not contain `recvmmsg`.
+#[cfg(not(any(target_os = "macos", target_os = "ios")))]
+unsafe fn recvmmsg_with_fallback(
+    sockfd: libc::c_int,
+    msgvec: *mut libc::mmsghdr,
+    vlen: libc::c_uint,
+) -> libc::c_int {
+    let flags = 0;
+    let timeout = ptr::null_mut::<libc::timespec>();
+
+    #[cfg(not(target_os = "freebsd"))]
+    {
+        let ret =
+            libc::syscall(libc::SYS_recvmmsg, sockfd, msgvec, vlen, flags, timeout) as libc::c_int;
+        if ret != -1 {
+            return ret;
+        }
+    }
+
+    // libc on FreeBSD implements `recvmmsg` as a high-level abstraction over `recvmsg`,
+    // thus `SYS_recvmmsg` constant and direct system call do not exist
+    #[cfg(target_os = "freebsd")]
+    {
+        let ret = libc::recvmmsg(sockfd, msgvec, vlen as usize, flags, timeout) as libc::c_int;
+        if ret != -1 {
+            return ret;
+        }
+    }
+
+    let e = io::Error::last_os_error();
+    match e.raw_os_error() {
+        Some(libc::ENOSYS) => {
+            // Fallback to `recvmsg`.
+            recvmmsg_fallback(sockfd, msgvec, vlen)
+        }
+        _ => -1,
+    }
+}
+
+/// Fallback implementation of `recvmmsg` using `recvmsg`
+/// for systems which do not support `recvmmsg`
+/// such as Linux <2.6.33.
+#[cfg(not(any(target_os = "macos", target_os = "ios")))]
+unsafe fn recvmmsg_fallback(
+    sockfd: libc::c_int,
+    msgvec: *mut libc::mmsghdr,
+    vlen: libc::c_uint,
+) -> libc::c_int {
+    let flags = 0;
+    if vlen == 0 {
+        return 0;
+    }
+
+    let n = libc::recvmsg(sockfd, &mut (*msgvec).msg_hdr, flags);
+    if n == -1 {
+        -1
+    } else {
+        // type of `msg_len` field differs on Linux and FreeBSD,
+        // it is up to the compiler to infer and cast `n` to correct type
+        (*msgvec).msg_len = n as _;
+        1
+    }
+}
+
 /// Returns the platforms UDP socket capabilities
-pub fn udp_state() -> UdpState {
+pub(crate) fn udp_state() -> UdpState {
     UdpState {
         max_gso_segments: AtomicUsize::new(gso::max_gso_segments()),
         gro_segments: gro::gro_segments(),
+        sendmsg_einval: AtomicBool::new(false),
     }
 }
 
@@ -370,6 +524,7 @@ fn prepare_msg(
     ctrl: &mut cmsg::Aligned<[u8; CMSG_LEN]>,
     #[allow(unused_variables)] // only used on FreeBSD & macOS
     encode_src_ip: bool,
+    sendmsg_einval: bool,
 ) {
     iov.iov_base = transmit.contents.as_ptr() as *const _ as *mut _;
     iov.iov_len = transmit.contents.len();
@@ -391,7 +546,9 @@ fn prepare_msg(
     let mut encoder = unsafe { cmsg::Encoder::new(hdr) };
     let ecn = transmit.ecn.map_or(0, |x| x as libc::c_int);
     if transmit.destination.is_ipv4() {
-        encoder.push(libc::IPPROTO_IP, libc::IP_TOS, ecn as IpTosTy);
+        if !sendmsg_einval {
+            encoder.push(libc::IPPROTO_IP, libc::IP_TOS, ecn as IpTosTy);
+        }
     } else {
         encoder.push(libc::IPPROTO_IPV6, libc::IPV6_TCLASS, ecn);
     }
@@ -475,6 +632,7 @@ fn decode_recv(
             (libc::IPPROTO_IPV6, libc::IPV6_TCLASS) => unsafe {
                 // Temporary hack around broken macos ABI. Remove once upstream fixes it.
                 // https://bugreport.apple.com/web/?problemID=48761855
+                #[allow(clippy::unnecessary_cast)] // cmsg.cmsg_len defined as size_t
                 if cfg!(target_os = "macos")
                     && cmsg.cmsg_len as usize == libc::CMSG_LEN(mem::size_of::<u8>() as _) as usize
                 {
@@ -542,10 +700,15 @@ fn decode_recv(
 
 #[cfg(not(any(target_os = "macos", target_os = "ios")))]
 // Chosen somewhat arbitrarily; might benefit from additional tuning.
-pub const BATCH_SIZE: usize = 32;
+pub(crate) const BATCH_SIZE: usize = 32;
 
 #[cfg(any(target_os = "macos", target_os = "ios"))]
-pub const BATCH_SIZE: usize = 1;
+pub(crate) const BATCH_SIZE: usize = 1;
+
+#[inline]
+pub(crate) fn may_fragment() -> bool {
+    false
+}
 
 #[cfg(target_os = "linux")]
 mod gso {
@@ -553,7 +716,7 @@ mod gso {
 
     /// Checks whether GSO support is available by setting the UDP_SEGMENT
     /// option on a socket
-    pub fn max_gso_segments() -> usize {
+    pub(crate) fn max_gso_segments() -> usize {
         const GSO_SIZE: libc::c_int = 1500;
 
         let socket = match std::net::UdpSocket::bind("[::]:0")
@@ -571,7 +734,7 @@ mod gso {
         }
     }
 
-    pub fn set_segment_size(encoder: &mut cmsg::Encoder, segment_size: u16) {
+    pub(crate) fn set_segment_size(encoder: &mut cmsg::Encoder, segment_size: u16) {
         encoder.push(libc::SOL_UDP, libc::UDP_SEGMENT, segment_size);
     }
 }
@@ -580,11 +743,11 @@ mod gso {
 mod gso {
     use super::*;
 
-    pub fn max_gso_segments() -> usize {
+    pub(super) fn max_gso_segments() -> usize {
         1
     }
 
-    pub fn set_segment_size(_encoder: &mut cmsg::Encoder, _segment_size: u16) {
+    pub(super) fn set_segment_size(_encoder: &mut cmsg::Encoder, _segment_size: u16) {
         panic!("Setting a segment size is not supported on current platform");
     }
 }
@@ -593,7 +756,7 @@ mod gso {
 mod gro {
     use super::*;
 
-    pub fn gro_segments() -> usize {
+    pub(crate) fn gro_segments() -> usize {
         let socket = match std::net::UdpSocket::bind("[::]:0")
             .or_else(|_| std::net::UdpSocket::bind("127.0.0.1:0"))
         {
@@ -641,7 +804,7 @@ const OPTION_ON: libc::c_int = 1;
 
 #[cfg(not(target_os = "linux"))]
 mod gro {
-    pub fn gro_segments() -> usize {
+    pub(super) fn gro_segments() -> usize {
         1
     }
 }
